/* descr.c - configuration description */

/* Written 1993 by Werner Almesberger */


#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "scanner.h"
#include "values.h"
#include "descr.h"


ITEM *descr = NULL;

static ERROR *errors = NULL,**next_error = &errors;
static ERROR *last_error = NULL;
static int error_ttl;

static int eval(const EXPR *expr);
static ITEM *item(ITEM *parent,ITEM *last,ITEM **here,int named);
static const ITEM *find_list_item(const ITEM *curr,const char *name);
static const EXPR *expr(void);


static int get_var(const ITEM *var,int value)
{
    if (value == -1) return VALUE_OF(var);
    return var->type == it_choice ? VALUE_OF(var) == value :
      VALUE_OF(var) & (1 << value) ? 1 : 0;
}


static int eval(const EXPR *expr)
{
    switch (expr->type) {
	case et_or:
	    return eval(expr->u.op.left) ? 1 : eval(expr->u.op.right);
	case et_xor:
	    return eval(expr->u.op.left) ? !eval(expr->u.op.right) :
	      eval(expr->u.op.right);
	case et_and:
	    return eval(expr->u.op.left) ? eval(expr->u.op.right) : 0;
	case et_not:
	    return !eval(expr->u.op.left);
	case et_uminus:
	    return -eval(expr->u.op.left);
	case et_plus:
	    return eval(expr->u.op.left)+eval(expr->u.op.right);
	case et_minus:
	    return eval(expr->u.op.left)-eval(expr->u.op.right);
	case et_times:
	    return eval(expr->u.op.left)*eval(expr->u.op.right);
	case et_div:
	    return eval(expr->u.op.left)/eval(expr->u.op.right);
	case et_mod:
	    return eval(expr->u.op.left)%eval(expr->u.op.right);
	case et_eq:
	    return eval(expr->u.op.left) == eval(expr->u.op.right);
	case et_ne:
	    return eval(expr->u.op.left) != eval(expr->u.op.right);
	case et_lt:
	    return eval(expr->u.op.left) < eval(expr->u.op.right);
	case et_gt:
	    return eval(expr->u.op.left) > eval(expr->u.op.right);
	case et_le:
	    return eval(expr->u.op.left) <= eval(expr->u.op.right);
	case et_ge:
	    return eval(expr->u.op.left) >= eval(expr->u.op.right);
	case et_atom:
	    return get_var(expr->u.atom.var,expr->u.atom.value);
	case et_number:
	    return expr->u.number.value;
	default:
	    abort();
    }
}


const char *dsc_error(void)
{
    ERROR *walk;

    for (walk = errors; walk; walk = walk->next)
	if (!walk->ignore && eval(walk->expr)) {
	    last_error = walk;
	    error_ttl = 2;
	    return walk->msg;
	}
    last_error = NULL;
    return NULL;
}


int dsc_ignore(void)
{
    if (last_error) last_error->ignore = 1;
    return !!last_error;
}


void dsc_age(void)
{
    if (!last_error) return;
    if (!--error_ttl) last_error = 0;
}


static const ITEM *find_item(const ITEM *curr,const char *name)
{
    const ITEM *this;
    int i;

    if (curr->name && !strcmp(curr->name,name))
	if (!IS_VARIABLE(curr)) scan_error("invalid item type");
	else return curr;
    switch (curr->type) {
	case it_choice:
	    for (i = 0; i < curr->num_values; i++)
		if (this = find_list_item(curr->values[i].entries,name))
		    return this;
		return NULL;
	case it_set:
	    return NULL;
	default:
	    return find_list_item(curr->entries,name);
    }
}


static const ITEM *find_list_item(const ITEM *list,const char *name)
{
    const ITEM *walk,*this;

    for (walk = list; walk; walk = walk->next)
	if (this = find_item(walk,name)) return this;
    return NULL;
}


static const EXPR *atom(void)
{
    const ITEM *item;
    const EXPR *tmp;
    EXPR *new;
    char *end;
    int i;

    scan_real();
    if (token != '(') scan_return();
    else {
	tmp = expr();
	scan_real();
	if (token != ')') scan_error(") expected");
	return tmp;
    }
    new = alloc(sizeof(EXPR));
    scan_string();
    new->u.number.value = strtol(string,&end,10);
    if (!*end) {
	new->type = et_number;
	return new;
    }
    new->type = et_atom;
    if (!(item = find_list_item(descr,string)))
	scan_error("no such item \"%s\"",string);
    new->u.atom.var = item;
    new->u.atom.value = -1;
    if (!scan_next()) {
	if (item->type == it_choice) scan_error("value suffix expected");
	return new;
    }
    if (token != '.') {
	if (item->type == it_choice) scan_error("value suffix expected");
	scan_return();
	return new;
    }
    scan_string();
    if (item->type == it_yesno)
	scan_error("unexpected value suffix \"%s\"",string);
    for (i = 0; i < item->num_values; i++)
	if (!strcmp(item->values[i].long_name,string) || (item->values[i].
	  short_name && !strcmp(item->values[i].short_name,string))) break;
    if (i == item->num_values) scan_error("no such value \"%s\"",string);
    new->u.atom.value = i;
    return new;
}


static const EXPR *unary_expr(void)
{
    EXPR *new;

    scan_real();
    if (token != '!' && token != '-') {
	scan_return();
	return atom();
    }
    new = alloc(sizeof(EXPR));
    new->type = token == '!' ? et_not : et_uminus;
    new->u.op.left = atom();
    new->u.op.right = NULL;
    return new;
}


static const EXPR *mult_expr(void)
{
    const EXPR *tmp;
    EXPR *new;

    tmp = unary_expr();
    if (!scan_next()) return tmp;
    if (token != '*' && token != '/' && token != '%') {
	scan_return();
	return tmp;
    }
    new = alloc(sizeof(EXPR));
    new->type = token == '*' ? et_plus : token == '/' ? et_div : et_mod;
    new->u.op.left = tmp;
    new->u.op.right = mult_expr();
    return new;
}


static const EXPR *add_expr(void)
{
    const EXPR *tmp;
    EXPR *new;

    tmp = mult_expr();
    if (!scan_next()) return tmp;
    if (token != '+' && token != '-') {
	scan_return();
	return tmp;
    }
    new = alloc(sizeof(EXPR));
    new->type = token == '+' ? et_plus : et_minus;
    new->u.op.left = tmp;
    new->u.op.right = add_expr();
    return new;
}


static const EXPR *rel_expr(void)
{
    const EXPR *tmp;
    EXPR *new;
    EXPR_TYPE type;

    tmp = add_expr();
    if (!scan_next()) return tmp;
    switch (token) {
	case '<':
	    type = et_lt;
	    scan_real();
	    if (token == '=') type = et_le;
	    else scan_return();
	    break;
	case '>':
	    type = et_gt;
	    scan_real();
	    if (token == '=') type = et_ge;
	    else scan_return();
	    break;
	case '=':
	    type = et_eq;
	    scan_real();
	    if (token != '=') scan_error("2nd = of == expected");
	    break;
	case '!':
	    type = et_ne;
	    scan_real();
	    if (token != '=') scan_error("= of != expected");
	    break;
	default:
	    scan_return();
	    return tmp;
    }
    new = alloc(sizeof(EXPR));
    new->type = type;
    new->u.op.left = tmp;
    new->u.op.right = rel_expr();
    return new;
}


static const EXPR *and_expr(void)
{
    const EXPR *tmp;
    EXPR *new;

    tmp = rel_expr();
    if (!scan_next()) return tmp;
    if (token != '&') {
	scan_return();
	return tmp;
    }
    scan_real();
    if (token != '&') scan_error("2nd & of && expected");
    new = alloc(sizeof(EXPR));
    new->type = et_and;
    new->u.op.left = tmp;
    new->u.op.right = and_expr();
    return new;
}


static const EXPR *expr(void)
{
    const EXPR *tmp;
    EXPR *new;

    tmp = and_expr();
    if (!scan_next()) return tmp;
    if (token != '|' && token != '^') {
	scan_return();
	return tmp;
    }
    new = alloc(sizeof(EXPR));
    new->type = token == '|' ? et_or : et_xor;
    if (token == '|') {
	scan_real();
	if (token != '|') scan_error("2nd | of || expected");
    }
    new->u.op.left = tmp;
    new->u.op.right = expr();
    return new;
}


static void error(void)
{
    ERROR *new;

    new = alloc(sizeof(ERROR));
    scan_string();
    new->msg = stralloc(string);
    new->expr = expr();
    new->ignore = 0;
    new->next = NULL;
    *next_error = new;
    next_error = &new->next;
}


static void name(const char **long_name,const char **short_name)
{
    scan_string();
    *long_name = stralloc(string);
    if (!scan_next()) return;
    if (token != ':') {
	scan_return();
	*short_name = NULL;
    }
    else {
	scan_string();
	*short_name = stralloc(string);
    }
}


static ITEM *create_item(ITEM_TYPE type,ITEM *parent,ITEM *prev,ITEM **here,
  int named)
{
    ITEM *new,*walk;

    new = alloc(sizeof(ITEM));
    new->type = type;
    new->name = new->long_name = new->short_name = NULL;
    new->deflt = type == it_choice ? -1 : 0;
    new->value = -1;
    new->active = 0;
    new->values = NULL;
    new->num_values = new->expand = 0;
    new->parent = parent;
    new->entries = new->next = NULL;
    new->prev = prev;
    if (named && type != it_title) {
	scan_string();
	new->name = stralloc(string);
    }
    name(&new->long_name,&new->short_name);
    for (walk = prev; walk; walk = walk->prev)
	if (!strcmp(walk->long_name,new->long_name))
	    scan_error("duplicate entry");
    *here = new;
    if (!named || !scan_next()) return new;
    if (token != '>') scan_return();
    else {
	scan_string();
	new->file = stralloc(string);
    }
    return new;
}


static void section(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    if (!scan_next()) return;
    if (token != '{') {
	scan_return();
	return;
    }
    while (1) {
	scan_real();
	if (token == '}') break;
	scan_return();
	last = item(parent,last,here,named);
	here = &last->next;
    }
}


static ITEM *clone(ITEM *orig)
{
    scan_error("not yet implemented");
    return NULL;
}


static void one_value(ITEM *parent,int nth)
{
    const char *long_name,*short_name,*value,*like;
    ITEM *entries;
    int i;

    scan_real();
    if (token != '*') scan_return();
    else if (parent->type == it_choice)
	    if (parent->deflt == -1) parent->deflt = nth;
	    else scan_error("duplicate default marker");
	else parent->deflt |= 1 << nth;
    value = NULL;
    entries = NULL;
    name(&long_name,&short_name);
    scan_real();
    if (token == '=') {
	scan_string();
	value = stralloc(string);
	scan_real();
    }
    like = NULL;
    if (token == '{') {
	scan_return();
	section(parent,NULL,&entries,parent->type == it_choice);
	scan_real();
    }
    else if (!token && strcmp(string,"like")) {
	    scan_string();
	    like = stralloc(string);
	}
    if (token == ',') one_value(parent,nth+1);
    else if (token != ')') scan_error(") expected");
	else {
	    parent->num_values = nth+1;
	    parent->values = alloc(sizeof(VALUE_ENTRY)*(nth+1));
	}
    parent->values[nth].long_name = long_name;
    parent->values[nth].short_name = short_name;
    parent->values[nth].value = value;
    parent->values[nth].expand = 0;
    parent->values[nth].entries = entries;
    for (i = 0; i < nth; i++)
	if (!strcmp(parent->values[i].long_name,parent->values[nth].long_name))
	    scan_error("duplicate value name");
    if (like) {
	for (i = 0; i < nth; i++)
	    if (!strcmp(parent->values[i].long_name,parent->values[nth].
	      long_name)) break;
	if (i == nth) scan_error("no matching value for \"like\"");
	if (parent->values[i].entries)
	    parent->values[nth].entries = parent->type == it_choice ?
	      parent->values[i].entries : clone(parent->values[i].entries);
	free((void *) like);
    }
}


static void value_list(ITEM *parent,int named)
{
    scan_real();
    if (token != '(') scan_error("( expected");
    one_value(parent,0);
    if (parent->type == it_choice && parent->deflt == -1)
	scan_error("default value is missing");
}


static ITEM *title(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    ITEM *item;

    item = create_item(it_title,parent,last,here,named);
    section(item,NULL,&item->entries,named);
    return item;
}


static ITEM *yesno(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    ITEM *item;

    item = create_item(it_yesno,parent,last,here,named);
    if (!scan_next()) return item;
    if (token != '*') scan_return();
    else item->deflt = 1;
    section(item,NULL,&item->entries,named);
    return item;
}


static int get_number(void)
{
    int value;
    char *end;

    value = strtol(string,&end,10);
    if (*end) scan_error("number expected");
    if (value < 0) scan_error("non-negative number expected");
    return value;
}


static ITEM *number(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    ITEM *item;

    item = create_item(it_number,parent,last,here,named);
    scan_real();
    if (token != '(') scan_error("( expected");
    scan_real();
    item->deflt = get_number();
    scan_real();
    if (token != ':') item->low = item->deflt;
    else {
	scan_real();
	item->low = get_number();
	scan_real();
    }
    if (token != ',') scan_error(", expected");
    scan_real();
    item->high = get_number();
    scan_real();
    if (token != ')') scan_error(") expected");
    section(item,NULL,&item->entries,named);
    return item;
}


static ITEM *choice(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    ITEM *item;

    item = create_item(it_choice,parent,last,here,named);
    value_list(item,named);
    return item;
}


static ITEM *set(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    ITEM *item;

    item = create_item(it_set,parent,last,here,named);
    value_list(item,named);
    return item;
}


static ITEM *extension(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    ITEM *walk,*scan;
    int i;

    scan_string();
    if (!last) walk = *here;
    else for (walk = last; walk->prev; walk = walk->prev);
    while (walk) {
	if (!strcmp(walk->long_name,string)) {
	    if (walk->type != it_choice && walk->type != it_set) {
		if (!walk->entries) scan = NULL;
		else for (scan = walk->entries; scan->next; scan = scan->next);
		section(walk,scan,scan ? &scan->next : &walk->entries,named);
		return last;
	    }
	    scan_real();
	    if (token != '.') scan_error(". expected");
	    scan_string();
	    for (i = 0; i < walk->num_values; i++)
		if (!strcmp(walk->values[i].long_name,string)) break;
	    if (i == walk->num_values) scan_error("unknown variable");
	    if (!walk->values[i].entries) scan = NULL;
	    else for (scan = walk->values[i].entries; scan->next; scan =
		  scan->next);
	    section(walk,scan,scan ? &scan->next : &walk->values[i].entries,
	      walk->type == it_choice && named);
	    return last;
	}
	walk = walk->next;
    }
    scan_error("unknown section");
    return NULL; /* for gcc */
}


static ITEM *item(ITEM *parent,ITEM *last,ITEM **here,int named)
{
    scan_string();
    if (!strcmp(string,"title")) return title(parent,last,here,named);
    if (!strcmp(string,"yesno")) return yesno(parent,last,here,named);
    if (!strcmp(string,"number")) return number(parent,last,here,named);
    if (!strcmp(string,"choice")) return choice(parent,last,here,named);
    if (!strcmp(string,"set")) return set(parent,last,here,named);
    if (!strcmp(string,"extension")) return extension(parent,last,here,named);
    scan_error("keyword expected");
    return NULL; /* to make gcc happy */
}


static void description(void)
{
    ITEM *last,**here;

    last = NULL;
    here = &descr;
    while (*here) here = &(last = *here)->next;
    while (scan_next()) {
	if (token) scan_error("string expected");
	if (!strcmp(string,"error")) error();
	else {
	    scan_return();
	    last = item(NULL,last,here,1);
	    here = &last->next;
	}
    }
}


void dsc_load(const char *name)
{
    scan_open(name);
    description();
    scan_close();
	
}
