/* file.c - output file handling */

/* Written 1993 by Werner Almesberger */


#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>

#include "common.h"
#include "scanner.h"
#include "file.h"


#define MAX_ERROR 500


char file_error[MAX_ERROR+1];

static char *templ_name = NULL,*tmp_name = NULL;


void file_reset(void)
{
    *file_error = 0;
    if (!templ_name) templ_name = alloc((size_t) pathconf("/",_PC_PATH_MAX)+1);
    if (!tmp_name) tmp_name = alloc((size_t) pathconf("/",_PC_PATH_MAX)+1);
}


void file_open(FILE_DSC *dsc,const char *name)
{
    const char *first,*dot;
    int ch;

    if (first = strrchr(name,'/')) first++;
    else first = name;
    dot = strrchr(name,'.');
    if (dot && dot <= first) dot = NULL;
    if (!dot) sprintf(templ_name,"%s.template",name);
    else {
	strncpy(templ_name,name,(size_t) (dot-name+1));
	strcpy(templ_name+(dot-name)+1,"template");
    }
    strncpy(tmp_name,name,(size_t) (first-name));
    tmp_name[first-name] = '~';
    strcpy(tmp_name+(first-name)+1,first);
    dsc->file_name = stralloc(name);
    dsc->templ_name = stralloc(templ_name);
    dsc->tmp_name = stralloc(tmp_name);
    dsc->templ_name = NULL;
    if (!(dsc->tmp = fopen(tmp_name,"w"))) {
	sprintf(file_error,"open %s: %s",tmp_name,strerror(errno));
	return;
    }
    if (dsc->templ = fopen(templ_name,"r")) {
	while ((ch = fgetc(dsc->templ)) != EOF)
	    if (ch == '$') break;
	    else if (fputc(ch,dsc->tmp) == EOF) {
		    sprintf(file_error,"write %s: %s",tmp_name,strerror(errno));
		    return;
		}
	if (ch == EOF) {
	    (void) fclose(dsc->templ);
	    dsc->templ = NULL;
	}
    }
}


const char *file_name(FILE_DSC *dsc)
{
    return dsc->file_name;
}


void file_printf(FILE_DSC *dsc,const char *fmt,...)
{
    va_list ap;

    if (*file_error) return;
    va_start(ap,fmt);
    if (vfprintf(dsc->tmp,fmt,ap) == EOF)
	sprintf(file_error,"write %s: %s",dsc->tmp_name,strerror(errno));
    va_end(ap);
}


void file_token(FILE_DSC *dsc,const char *token)
{
    const char *walk;

    if (*file_error) return;
    if (!scan_quote(token)) {
	file_printf(dsc,"%s",token);
	return;
    }
    if (fputc('"',dsc->tmp) != EOF) {
	for (walk = token; *walk; walk++) {
	    if (*walk == '"' || *walk == '\\')
		if (fputc('\\',dsc->tmp) == EOF) break;
	    if (fputc(*walk,dsc->tmp) == EOF) break;
	}
	if (!*walk && fputc('"',dsc->tmp) != EOF) return;
    }
    sprintf(file_error,"write %s: %s",dsc->tmp_name,strerror(errno));
}


void file_close(FILE_DSC *dsc)
{
    int ch;

    if (!*file_error)
	while ((ch = fgetc(dsc->templ)) != EOF)
	    if (fputc(ch,dsc->tmp) == EOF) {
		sprintf(file_error,"write %s: %s",dsc->tmp_name,
		  strerror(errno));
		break;
	    }
    if (dsc->templ) (void) fclose(dsc->templ);
    if (fclose(dsc->tmp) == EOF && !*file_error)
	sprintf(file_error,"close %s: %s",dsc->tmp_name,strerror(errno));
    dsc->templ = dsc->tmp = NULL;
    free((char *) dsc->templ_name);
    dsc->templ_name = NULL;
}


static int differ(const char *old_name,const char *new_name)
{
    FILE *old,*new;
    int old_ch,new_ch;

    if (!(old = fopen(old_name,"r"))) return 1;
    if (!(new = fopen(new_name,"r"))) {
	sprintf(file_error,"re-open %s: %s",new_name,strerror(errno));
	return 0;
    }
    while ((old_ch = fgetc(old)) == (new_ch = fgetc(new)))
	if (old_ch == EOF) break;
    (void) fclose(old);
    (void) fclose(new);
    return old_ch != new_ch;
}


void file_done(FILE_DSC *dsc,int commit)
{
    if (!commit || !differ(dsc->file_name,dsc->tmp_name))
	(void) remove(dsc->tmp_name);
    else if (rename(dsc->tmp_name,dsc->file_name) < 0 && !*file_error)
	    sprintf(file_error,"rename %s to %s: %s",dsc->tmp_name,
	      dsc->file_name,strerror(errno));
    free((char *) dsc->tmp_name);
    free((char *) dsc->file_name);
}
