/* full.c - full-screen configuration editor */

/* Written 1993 by Werner Almesberger */


#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "common.h"
#include "cursed.h"
#include "kconf.h"
#include "descr.h"
#include "tree.h"
#include "output.h"


#define VALUES_OFFSET 3
#define VALUES_GAP 1
#define VALUES_INDENT 1
#define RIGHT_ALIGN
#ifdef RIGHT_ALIGN
#define RIGHT_MARGIN 2
#else
#define RIGHT_MARGIN 0
#endif

#define STATUS_LINE (area_end+4)
#define YES "yY\r\n "
#define NO "nN\010\177"

typedef struct {
    int x,y;
} XY;

static int area_top,area_end,x_width;


static void frame(void)
{
    clear();
    move(0,0);
    standout();
    printw(" Linux Kernel Configurator V0.2 ");
    standend();
    move(area_end+2,2);
    printw("up/down to select item  left/right to change value  "
      "[+]/[-] to display/hide");
    move(area_end+3,11);
    printw("[Enter] to edit  [W] to write  [Q] to quit without saving");
}


static XY *init_values(ITEM *curr,int title,int indent,int *rows)
{
    int col,i,item_len,ind,values,right,end;
    XY *pos;

    switch (curr->type) {
	case it_title:
	    *rows = 0;
	    return NULL;
	case it_yesno:
	    *rows = 0;
	    col = 9+VALUES_GAP; /* [Yes]<gap>[No] */
	    pos = alloc(sizeof(XY)*2);
	    pos[0].x = indent+VALUES_INDENT;
	    pos[1].x = indent+5+VALUES_INDENT+VALUES_GAP;
	    pos[0].y = pos[1].y = 0;
	    break;
	case it_number:
	    *rows = 0;
	    col = number_length(curr->high)+2; /* 2: square brackets */
	    pos = alloc(sizeof(XY));
	    pos[0].x = indent+VALUES_INDENT;
	    pos[0].y = 0;
	    break;
	case it_choice:
	case it_set:
	    *rows = col = 0;
	    pos = alloc(sizeof(XY)*curr->num_values);
	    for (i = 0; i < curr->num_values; i++) {
		item_len = strlen(tree_nth_value(curr,i))+2;
		ind = col ? VALUES_GAP : 0;
		if (indent+col+item_len+ind >= x_width) {
		    col = ind = 0;
		    (*rows)++;
		}
		pos[i].x = col+indent+VALUES_INDENT+ind;
		pos[i].y = *rows;
		col += item_len+ind;
	    }
	    break;
	default:
	    abort();
    }
    values = tree_num_values(curr);
    if (!*rows && col+indent+title+VALUES_OFFSET+RIGHT_MARGIN <= x_width)
        for (i = 0; i < values; i++)
#ifdef RIGHT_ALIGN
	    pos[i].x += x_width-col-indent-VALUES_INDENT-RIGHT_MARGIN;
#else
	    pos[i].x += title+VALUES_OFFSET-VALUES_INDENT;
#endif
    else {
	(*rows)++;
	right = 0;
	for (i = 0; i < values; i++) {
	    pos[i].y++;
	    end = pos[i].x+strlen(tree_nth_value(curr,i))+2-1;
	    if (end > right) right = end;
	}
#ifdef RIGHT_ALIGN
	for (i = 0; i < values; i++) pos[i].x += x_width-right-RIGHT_MARGIN;
#endif
    }
    return pos;
}


static void show_values(ITEM *curr,XY *pos,int y,int mark)
{
    int i,value,selected,marked;

    if (!IS_VARIABLE(curr)) return;
    for (i = 0; i < tree_num_values(curr); i++) {
	if (y+pos[i].y <= area_end) {
	    move(y+pos[i].y,pos[i].x);
	    value = curr->type == it_yesno ? 1-i : i;
	    switch (curr->type) {
		case it_yesno:
		    selected = marked = VALUE_OF(curr) ? value : !value;
		    break;
		case it_number:
		    selected = marked = 1;
		    break;
		case it_choice:
		    selected = marked = i == VALUE_OF(curr);
		    break;
		case it_set:
		    selected = VALUE_OF(curr) & (1 << i);
		    marked = curr->active == i;
		    break;
		default:
		    abort();
	    }
	    if (marked && mark) standout();
	    printw("%c%s%c",selected ? '[' : ' ',tree_nth_value(curr,value),
	      selected ? ']' : ' ');
	    if (marked && mark) standend();
	}
    }
}


static int indentation(ITEM *curr)
{
    return 2*tree_level(curr)+2;
}


static int show_entry(ITEM *curr,int y,int mark)
{
    int indent,rows;
    const char *line;
    XY *pos;

    indent = indentation(curr);
    if (tree_expanded(curr)) line = curr->long_name;
    else line = tree_line(curr,x_width-indent);
    pos = init_values(curr,(int) strlen(line),indent,&rows);
    move(y,0);
    printw(tree_new(curr) ? "*" : !tree_expanded(curr) && tree_section(curr) ?
      "+" : " ");
    move(y,indent);
    if (mark) standout();
    printw("%s",line);
    if (mark) standend();
    show_values(curr,pos,y,mark);
    if (pos) free(pos);
    return rows+1;
}


static int entry_size(ITEM *curr)
{
    const char *line;
    XY *pos;
    int rows;

    if (!IS_VARIABLE(curr)) return 1;
    if (tree_expanded(curr)) line = curr->long_name;
    else line = tree_line(curr,x_width-indentation(curr));
    pos = init_values(curr,(int) strlen(line),indentation(curr),&rows);
    free(pos);
    return rows+1;
}


static void clear_area(int y)
{
    int i;

    for (i = y; i <= area_end; i++) {
	move(i,0);
	clrtoeol();
    }
}


static void draw_area(ITEM *top,int y)
{
    int lines;

    clear_area(y);
    while (y <= area_end) {
	lines = show_entry(top,y,0);
	y += lines;
	if (!tree_next(&top)) break;
    }
}


static int sub_size(ITEM *curr)
{
    ITEM *walk;
    int size;

    size = entry_size(curr);
    if (tree_expanded(curr))
	for (walk = tree_entries(curr); walk; walk = walk->next)
	    size += sub_size(walk);
    return size;
}


static int ask(const char *question)
{
    int ch;

    move(STATUS_LINE,0);
    printw("%s ? (Y/N) ",question);
    refresh();
    while (1) {
	ch = getch();
	if (strchr(YES NO,ch)) break;
    }
    move(STATUS_LINE,0);
    clrtoeol();
    return !strchr(NO,ch);
}


static const char *limit(const char *str)
{
    static char *buffer = NULL;

    if (!buffer) buffer = alloc((size_t) (x_width+1));
    strncpy(buffer,str,(size_t) x_width);
    buffer[x_width] = 0;
    return buffer;
}


static void edit_number(ITEM *curr,int y)
{
    const char *line;
    XY *pos;
    int rows,value,width,done,ch;

    if (tree_expanded(curr)) line = curr->long_name;
    else line = tree_line(curr,x_width-indentation(curr));
    pos = init_values(curr,(int) strlen(line),indentation(curr),&rows);
    value = VALUE_OF(curr);
    width = number_length(curr->high);
    move(STATUS_LINE,0);
    clrtoeol();
    move(STATUS_LINE,0);
    standout();
    printw(" Range of valid values is %d..%d. Press [Q] to keep old value. ",
      curr->low,curr->high);
    standend();
    done = 0;
    do {
	move(y+pos[0].y,pos[0].x+1);
	printw("%*d",width,value);
	refresh();
	switch (ch = getch()) {
	    case '\b':
	    case 127:
		value /= 10;
		break;
	    case 'U'-64:
	    case 'X'-64:
		value = 0;
		break;
	    case '\r':
	    case '\n':
		if (value < curr->low) break;
		tree_set_number(curr,value);
		/* fall through */
	    case 'Q':
	    case 'q':
		done = 1;
		break;
	    default:
		if (!isdigit(ch)) break;
		ch -= '0';
		if (value > curr->high/10 || value*10 > curr->high-ch) break;
		value = value*10+ch;
	}
    }
    while (!done);
    move(STATUS_LINE,0);
    clrtoeol();
}


static void editor(ITEM *root)
{
    ITEM *curr,*top,*before,*section_before,*parent,*tmp;
    int y,ch,own_size,expanded_before,tmp_y;
    int status_line,redraw,esc;
    const char *error;

    status_line = redraw = 0;
    y = area_top;
    curr = top = root;
    while (1) {
	own_size = entry_size(curr);
	expanded_before = tree_expanded(curr);
	section_before = tree_section(curr);
	before = curr;
	if (y < area_top) {
	    y = area_top;
	    top = curr;
	    draw_area(top,y);
	}
	if (y+sub_size(curr)-1 > area_end && y != area_top) {
	    while (y+sub_size(curr)-1 > area_end && y != area_top) {
		y -= entry_size(top);
		tree_next(&top);
	    }
	    draw_area(top,area_top);
	    redraw = 0;
	}
	if (IS_VARIABLE(curr) && curr->value == -1) {
	    curr->value = curr->deflt == -1 ? 0 : curr->deflt;
	    if (curr->parent && !tree_new(curr->parent) && !no_new) redraw = 1;
	}
	if (redraw) {
	    tmp = curr;
	    tmp_y = y;
	    while (tmp_y > area_top) {
		assert(tree_prev(&tmp));
		tmp_y -= entry_size(tmp);
	    }
	    assert(tmp_y == area_top);
	    draw_area(tmp,area_top);
	    redraw = 0;
	}
	show_entry(curr,y,1);
	refresh();
	esc = 0;
	do {
	    ch = getch();
	    if (ch == 27 && !esc) esc = 1;
	    else if (!esc || (ch != '[' && ch !='O')) break;
	}
	while (1);
	if (!strchr("ABCD",ch)) esc = 0;
	show_entry(curr,y,0);
	if (status_line) {
	    move(STATUS_LINE,0);
	    clrtoeol();
	    status_line = 0;
	}
	dsc_age();
	switch (ch) {
	    case 'D':
		if (!esc) break;
	    case 'H':
	    case 'h':
	    case 'B'-32:
		tree_change(curr,0);
		break;
	    case 'C':
		if (!esc) break;
	    case 'L':
	    case 'l':
	    case 'F'-32:
	    case '\t':
		tree_change(curr,1);
		break;
	    case 'B':
		if (!esc) break;
	    case 'J':
	    case 'j':
	    case 'N'-32:
		if (tree_next(&curr)) y += own_size;
		continue;
	    case 'A':
		if (!esc) break;
	    case 'K':
	    case 'k':
	    case 'P'-32:
		if (tree_prev(&curr)) y -= entry_size(curr);
		continue;
	    case ' ':
		if (curr->type == it_set) tree_toggle(curr);
		else if (curr->type == it_number) edit_number(curr,y);
		    else if (IS_LISTABLE(curr) && (curr->type != it_yesno ||
			  !VALUE_OF(curr) || !tree_entries(curr)))
			    tree_change(curr,1);
			else EXPAND(curr) = !EXPAND(curr);
		break;
	    case '\n':
	    case '\r':
		if (curr->type == it_number) edit_number(curr,y);
		break;
	    case '+':
		if (EXPAND(curr)) break;
		EXPAND(curr) = 1;
		break;
	    case 127:
	    case 8:
	    case '-':
		if ((!EXPAND(curr) || !tree_section(curr)) && (parent =
		  curr->parent)) {
		    while (curr != parent) {
			assert(tree_prev(&curr));
			y -= entry_size(curr);
		    }
		}
		EXPAND(curr) = 0;
		break;
	    case 'I':
	    case 'i':
		if (!dsc_ignore()) break;
		/* fall through */
	    case 'W':
	    case 'w':
		if (error = dsc_error()) {
		    move(STATUS_LINE,0);
		    standout();
		    printw(strlen(error) <= x_width-2 ? " %s " : "%s",
		      limit(error));
		    standend();
		    status_line = 1;
		    break;
		}
		if (!ask("Write configuration")) break;
		move(STATUS_LINE,0);
		clrtoeol();
		error = out_tree();
		move(STATUS_LINE,0);
		printw("%s",error ? limit(error) : "Wrote new configuration");
		status_line = 1;
		break;
	    case 'Q':
	    case 'q':
		if (ask("Really quit")) return;
		break;
	    default:
		break;
	}
	if (tree_new(curr)) redraw = 1;
	else if (curr != before || (tree_expanded(curr) != expanded_before &&
	      tree_section(curr) || section_before) || tree_section(curr) !=
	      section_before) draw_area(curr,y);
    }
}


void full_screen(ITEM *root)
{
    initscr();
    area_top = 2;
    area_end = LINES-5;
    x_width = COLS;
    noecho();
    raw();
    frame();
    draw_area(root,area_top);
    editor(root);
    clear();
    move(0,0);
    refresh();
    echo();
    noraw();
    endwin();
}
