/* line.c - line-mode configuration editor */

/* Written 1993 by Werner Almesberger */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "descr.h"
#include "tree.h"
#include "output.h"


#define LINE_LEN 300
#define SCREEN_WIDTH 80


static void prompt(ITEM *item)
{
    if (item->type != it_set)
	printf("%s%s [%s] ",tree_new(item) ? "* " : "",item->long_name,
	  tree_value(item));
    else {
	if (item->value == -1) item->value = item->deflt;
	printf("%s%s \"%s\" ? [%s] ",tree_new(item) ? "* " : "",
	  item->long_name,item->values[item->active].long_name,
	  item->value & (1 << item->active) ? "Yes" : "No");
    }
}


static void write_prompt(ITEM *item)
{
    printf("Write configuration ? [Yes] ");
}


static void ignore_prompt(ITEM *item)
{
    printf("Ignore ? [No] ");
}


static const char *yesno(const ITEM *item,int index)
{
    return index ? "Yes" : "No";
}


static char *trim(char *line)
{
    char *here;

    for (here = strchr(line,0)-1; here >= line; here--)
	if (*here != ' ' && *here != '\t' && *here != '\n') break;
	else *here = 0;
    for (here = line; *here; here++)
	if (*here != ' ' && *here != '\t') break;
    return *here ? here : NULL;
}


static int get_input(ITEM *item,void (*prompt)(ITEM *item),
  const char *(*nth_value)(const ITEM *item,int index),int values)
{
    char line[LINE_LEN+2];
    char *here;
    const char *value;
    int this,i,col;

    while (1) {
	prompt(item);
	fflush(stdout);
	if (!fgets(line,LINE_LEN,stdin)) exit(0);
	if (!(here = trim(line))) return -1;
	if (!strcmp(here,"?")) {
	    printf("Valid values are:\n  ");
	    col = 0;
	    for (i = 0; i < values; i++) {
		value = nth_value(item,i);
		if (col && col+strlen(value)+2 > SCREEN_WIDTH-2) {
		    col = 0;
		    printf("\n  ");
		}
		printf("%s%s",col ? ", " : "",value);
		col += strlen(value)+(col ? 2 : 0);
	    }
	    printf("\n");
	    continue;
	}
	this = -1;
	for (i = 0; i < values; i++)
	    if (!strncasecmp(here,nth_value(item,i),strlen(here)))
		if (this == -1) this = i;
		else {
		    this = -1;
		    break;
		}
	if (this >= 0) return this;
	fprintf(stderr,"Invalid input\n");
    }
}


static void get_number(ITEM *item)
{
    char line[LINE_LEN+2];
    char *here,*end;
    int value;

    while (1) {
	printf("%s%s [%d] ",tree_new(item) ? "* " : "",item->long_name,
	  VALUE_OF(item));
	fflush(stdout);
	if (!fgets(line,LINE_LEN,stdin)) exit(0);
	if (!(here = trim(line))) return;
	if (!strcmp(here,"?")) {
	    printf("Range of valid values is %d..%d\n",item->low,item->high);
	    continue;
	}
	value = strtol(line,&end,10);
	if (!*end && value >= item->low && value <= item->high) {
	    tree_set_number(item,value);
	    return;
	}
	fprintf(stderr,"Invalid input\n");
    }
}


static void do_round(ITEM *root)
{
    int i,value;

    do {
	root->expand = 1;
	if (root->values)
	    for (i = 0; i < root->num_values; i++) root->values[i].expand = 1;
	if (tree_entries(root) || root->type == it_title)
	    printf("|\n|  %s\n|\n",root->long_name);
	if (root->type == it_title) continue;
	root->value = VALUE_OF(root);
	if (root->type == it_number) get_number(root);
	else if ((value = get_input(root,prompt,root->type == it_set ? yesno :
	      tree_nth_value,root->type == it_set ? 2 : tree_num_values(root)))
	      >= 0)
		if (root->type != it_set) root->value = value;
		else if (value) root->value |= 1 << root->active;
		    else root->value &= ~(1 << root->active);
    }
    while (tree_seq(&root));
}


void line_mode(ITEM *root)
{
    const char *error;

    root->active = -1;
    while (1) {
	do_round(root);
	putchar('\n');
        if (!get_input(NULL,write_prompt,yesno,2)) return;
	while (1) {
	    if (!(error = dsc_error())) break;
	    fprintf(stderr,"\n%s\n",error);
	    if (get_input(NULL,ignore_prompt,yesno,2) == 1) dsc_ignore();
	    else break;
	}
	if (!error) break;
    }
    if (error = out_tree()) fprintf(stderr,"\n%s\n",error);
    else fprintf(stderr,"\nWrote new configuration\n");
}
