/* tree.c - configuration data tree access */

/* Written 1993 by Werner Almesberger */


#include <stdlib.h>
#include <assert.h>

#include "kconf.h"
#include "common.h"
#include "file.h"
#include "descr.h"
#include "values.h"
#include "output.h"


typedef struct _file_entry {
    FILE_DSC dsc;
    struct _file_entry *next;
} FILE_ENTRY;

static FILE_ENTRY *files;
static FILE_DSC config,make;


static void write_tree(ITEM *curr,FILE_DSC *file,const char *curr_name,
  int enabled);


static void init_tree(ITEM *curr)
{
    int i;

    while (curr) {
	curr->written = 0;
	if (curr->entries) init_tree(curr->entries);
	if (curr->values)
	    for (i = 0; i < curr->num_values; i++)
		init_tree(curr->values[i].entries);
	curr = curr->next;
    }
}


static void write_path(ITEM *this,FILE_DSC *file)
{
    ITEM *parent;
    int i;

    while (this->prev) this = this->prev;
    if (!(parent = this->parent)) return;
    write_path(parent,file);
    if (parent->type == it_set) {
	for (i = 0; i < parent->num_values; i++)
	    if (parent->values[i].entries == this) break;
	assert(i < parent->num_values);
	file_token(file,parent->name ? parent->name : parent->long_name);
	file_printf(file,".");
	file_token(file,parent->values[i].long_name);
	file_printf(file,",");
    }
}


static void write_named(ITEM *curr,FILE_DSC *file,const char *name,int enabled)
{
    int value,i,count,first;
    const char *val;

    value = VALUE_OF(curr);
    switch (curr->type) {
	case it_yesno:
	    if (curr->value != -1)
		file_printf(&config,"%s%s\n",curr->name,value ? "" : "!");
	    if (enabled)
		if (value) file_printf(&make,"%s = %s\n",curr->name,curr->name);
		else file_printf(&make,"# %s is not set\n",curr->name);
	    if (value && enabled && file)
		file_printf(file,"#define %s 1\n",curr->name);
	    break;
	case it_number:
	    if (curr->value != -1)
		file_printf(&config,"%s=%d\n",curr->name,value);
	    if (enabled) {
		file_printf(&make,"%s = %d\n",curr->name,value);
		if (file) file_printf(file,"#define %s %d\n",curr->name,value);
	    }
	    break;
	case it_choice:
	    val = curr->values[value].value ? curr->values[value].value :
	      curr->values[value].long_name;
	    if (curr->value != -1) {
		file_printf(&config,"%s=",curr->name);
		file_token(&config,curr->values[value].long_name);
		file_printf(&config,"\n");
	    }
	    if (enabled) {
		file_printf(&make,"%s = %s\n",curr->name,val);
		if (file) file_printf(file,"#define %s %s\n",curr->name,val);
	    }
	    break;
	case it_set:
	    if (curr->value != -1) {
		file_printf(&config,"%s(",curr->name);
		first = 1;
		for (i = 0; i < curr->num_values; i++)
		    if (value & (1 << i)) {
			if (!first) file_printf(&config,",");
			first = 0;
			file_token(&config,curr->values[i].long_name);
		    }
		file_printf(&config,")\n");
	    }
	    count = 0;
	    for (i = 0; i < curr->num_values; i++)
		if (value & (1 << i)) count++;
	    if (enabled) {
		if (count) file_printf(&make,"%s = %d\n",curr->name,count);
		else file_printf(&make,"# %s = 0\n",curr->name);
		if (file) file_printf(file,"#define %s { %d",curr->name,count);
	    }
	    for (i = 0; i < curr->num_values; i++)
		if (!(enabled && (value & (1 << i))))
		    write_tree(curr->values[i].entries,file,name,0);
		else {
		    if (file)
			file_printf(file,", \\\n  %s",curr->values[i].value ?
			  curr->values[i].value : curr->values[i].long_name);
		    write_tree(curr->values[i].entries,file,name,1);
		}
	    if (enabled && file) file_printf(file," }\n");
	default:
	    break;
    }
}


static void write_unnamed(ITEM *curr,FILE_DSC *file,const char *name,
  int enabled)
{
    int value,first,i,count;
    const char *val;

    value = VALUE_OF(curr);
    switch (curr->type) {
	case it_yesno:
	    if (curr->value != -1) {
		write_path(curr,&config);
		file_token(&config,curr->long_name);
		file_printf(&config,"%s\n",value ? "" : "!");
	    }
	    if (enabled && file) file_printf(file,",%d",value);
	    break;
	case it_number:
	    if (curr->value != -1) {
		write_path(curr,&config);
		file_token(&config,curr->long_name);
		file_printf(&config,"=%d\n",value);
	    }
	    if (enabled && file) file_printf(file,",%d",value);
	    break;
	case it_choice:
	    val = curr->values[value].value ? curr->values[value].value :
	      curr->values[value].long_name;
	    if (curr->value != -1) {
		write_path(curr,&config);
		file_token(&config,curr->long_name);
		file_printf(&config,"=");
		file_token(&config,curr->values[value].long_name);
		file_printf(&config,"\n");
	    }
	    if (enabled && file) file_printf(file,",%s",val);
	    break;
	case it_set:
	    if (curr->value != -1) {
		write_path(curr,&config);
		file_token(&config,curr->long_name);
		file_printf(&config,"(");
		first = 1;
		for (i = 0; i < curr->num_values; i++)
		    if (value & (1 << i)) {
			if (!first) file_printf(&config,",");
			first = 0;
			file_token(&config,curr->values[i].long_name);
		    }
		file_printf(&config,")\n");
	    }
	    if (enabled) {
		count = 0;
		for (i = 0; i < curr->num_values; i++)
		    if (value & (1 << i)) count++;
		if (file) file_printf(file,",%d",count);
	    }
	    for (i = 0; i < curr->num_values; i++)
		if (!(enabled && (value & (1 << i))))
		    write_tree(curr->values[i].entries,file,name,0);
		else {
		    if (file)
			file_printf(file,", \\\n    %s",curr->values[i].value ?
			  curr->values[i].value : curr->values[i].long_name);
		    write_tree(curr->values[i].entries,file,name,1);
		}
	default:
	    break;
    }
}


static void write_tree(ITEM *curr,FILE_DSC *file,const char *curr_name,
  int enabled)
{
    const char *name;
    int i;

    while (curr) {
	name = curr->file ? curr->file : curr_name;
	if (IS_VARIABLE(curr) && !curr->written &&
	  (file ? name && !strcmp(file_name(file),name) : !name)) {
	    curr->written = 1;
	    if (curr->name) write_named(curr,file,name,enabled);
	    else write_unnamed(curr,file,name,enabled);
	}
	else {
	    if (curr->type == it_set)
		for (i = 0; i < curr->num_values; i++)
		    write_tree(curr->values[i].entries,file,name,enabled &&
		      (VALUE_OF(curr) & (1 << i)));
	}
	switch (curr->type) {
	    case it_title:
	    case it_number:
		write_tree(curr->entries,file,name,enabled);
		break;
	    case it_yesno:
		write_tree(curr->entries,file,name,enabled && VALUE_OF(curr));
		break;
	    case it_choice:
		for (i = 0; i < curr->num_values; i++)
		    write_tree(curr->values[i].entries,file,name,
		      enabled && i == VALUE_OF(curr));
		break;
	    case it_set:
		/* handled above or in write_[un]named */
		break;
	    default:
		abort();
	}
	curr = curr->next;
    }
}


static void find_files(ITEM *curr,const char *curr_name)
{
    const char *name;
    FILE_ENTRY *new;
    int i;

    while (curr) {
	name = curr->file ? curr->file : curr_name;
	if (name && IS_VARIABLE(curr) && !curr->written) {
	    if (*file_error) return; /* no need to waste time */
	    new = alloc(sizeof(FILE_ENTRY));
	    file_open(&new->dsc,name);
	    new->next = files;
	    files = new;
	    write_tree(descr,&new->dsc,default_output,1);
	    file_close(&new->dsc);
	}
	find_files(curr->entries,name);
	if (curr->values)
	    for (i = 0; i < curr->num_values; i++)
		find_files(curr->values[i].entries,name);
	curr = curr->next;
    }
}


const char *out_tree(void)
{
    FILE_ENTRY *this;
    int okay;

    file_reset();
    files = NULL;
    file_open(&config,config_output);
    file_open(&make,make_output);
    init_tree(descr);
    if (!default_output) write_tree(descr,NULL,NULL,1);
	/* write variables not in a header file */
    find_files(descr,default_output);
    if (obsolete && *obsolete)
	file_printf(&config,"# The following variables have not been used by "
	  "the current configuration.\n# They may be obsolete. You may wish "
	  "to remove them.\n%s",obsolete);
    file_close(&config);
    file_close(&make);
    file_done(&config,!*file_error);
    okay = !*file_error;
    file_done(&make,okay);
    while (files) {
	file_done(&files->dsc,okay);
	this = files;
	files = files->next;
	free(this);
    }
    return *file_error ? file_error : NULL;
}
