/* tree.c - configuration data tree access */

/* Written 1993 by Werner Almesberger */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "common.h"
#include "descr.h"
#include "tree.h"
#include "kconf.h"


#define MAX_SCRATCH 2000
#define MAX_ERROR 300


int tree_num_values(const ITEM *curr)
{
    switch (curr->type) {
	case it_yesno:
	    return 2;
	case it_number:
	    return 1;
	case it_choice:
	case it_set:
	    return curr->num_values;
	default:
	    abort();
    }
}


const char *tree_nth_value(const ITEM *curr,int number)
{
    static char tmp[20]; /* sigh ... */

    switch (curr->type) {
	case it_yesno:
	    return number ? "Yes" : "No";
	case it_number:
	    sprintf(tmp,"%*d",number_length(curr->high),VALUE_OF(curr));
	    return tmp;
	case it_choice:
	case it_set:
	    return curr->values[number].long_name;
	default:
	    abort();
    }
}


const char *tree_value(const ITEM *curr)
{
    assert(curr->type != it_set);
    return tree_nth_value(curr,VALUE_OF(curr));
}


int tree_expanded(const ITEM *curr)
{
    switch (curr->type) {
	case it_title:
	    return curr->expand;
	case it_yesno:
	case it_number:
	    return VALUE_OF(curr) ? curr->expand : 0;
	case it_choice:
	    return curr->values[VALUE_OF(curr)].expand;
	case it_set:
	    if (!(VALUE_OF(curr) & (1 << curr->active))) return 0;
	    return curr->values[curr->active].expand;
	default:
	    return 1;
    }
}


ITEM *tree_section(const ITEM *curr)
{
    switch (curr->type) {
	case it_title:
	    return curr->entries;
	case it_yesno:
	case it_number:
	    return VALUE_OF(curr) ? curr->entries : NULL;
	case it_choice:
	    return curr->values[VALUE_OF(curr)].entries;
	case it_set:
	    if (!(VALUE_OF(curr) & (1 << curr->active))) return NULL;
	    return curr->values[curr->active].entries;
	default:
	    return NULL;
    }
}


ITEM *tree_entries(const ITEM *curr)
{
    switch (curr->type) {
	case it_title:
	case it_yesno:
	case it_number:
	    return curr->entries;
	case it_choice:
	    return curr->values[VALUE_OF(curr)].entries;
	case it_set:
	    return curr->values[curr->active].entries;
	default:
	    return NULL;
    }
}


int tree_prev(ITEM **curr)
{
    ITEM *walk;

    if ((*curr)->prev) {
	walk = (*curr)->prev;
	while (tree_entries(walk) && tree_expanded(walk))
	    for (walk = tree_entries(walk); walk->next; walk = walk->next);
	*curr = walk;
	return 1;
    }
    if ((*curr)->parent) {
	*curr = (*curr)->parent;
	return 1;
    }
    return 0;
}


static int tree_step(ITEM **curr,int seq)
{
    ITEM *walk;

    if (tree_entries(*curr) && tree_expanded(*curr)) {
	*curr = tree_entries(*curr);
	if (seq) (*curr)->active = 0;
	return 1;
    }
    if (seq && (*curr)->type == it_set && (*curr)->active < (*curr)->
      num_values-1) {
	(*curr)->active++;
	return 1;
    }
    if ((*curr)->next) {
	*curr = (*curr)->next;
	if (seq) (*curr)->active = 0;
	return 1;
    }
    for (walk = (*curr)->parent; walk; walk = walk->parent) {
	if (seq && walk->type == it_set && walk->active < walk->num_values-1) {
	    walk->active++;
	    *curr = walk;
	    return 1;
	}
	if (walk->next) {
	    *curr = walk->next;
	    if (seq) (*curr)->active = 0;
	    return 1;
	}
    }
    return 0;
}


int tree_next(ITEM **curr)
{
    return tree_step(curr,0);
}


int tree_seq(ITEM **curr)
{
    return tree_step(curr,1);
#if 0
    if ((*curr)->type != it_set || (*curr)->active >= (*curr)->num_values-1)
	return tree_step(curr,1);
    (*curr)->active++;
    return 1;
#endif
}


int tree_level(const ITEM *curr)
{
    ITEM *walk;
    int level;

    level = 0;
    for (walk = curr->parent; walk; walk = walk->parent) level++;
    return level;
}


int tree_new(const ITEM *curr)
{
    ITEM *walk;
    int i;

    if (no_new) return 0;
    if (IS_VARIABLE(curr) && curr->value == -1) return 1;
    if ((curr->type == it_yesno || curr->type == it_number) && !VALUE_OF(curr))
	return 0;
    if (curr->type != it_set) {
	for (walk = tree_entries(curr); walk; walk = walk->next)
	    if (tree_new(walk)) return 1;
    }
    else for (i = 0; i < curr->num_values; i++)
	    if (VALUE_OF(curr) & (1 << i))
		for (walk = curr->values[i].entries; walk; walk = walk->next)
		    if (tree_new(walk)) return 1;
    return 0;
}


static int sub_tree(const ITEM *curr,char *here,int top,int levels);


static int sub_list(const ITEM *curr,const ITEM *list,char *here,int top,
  int levels)
{
    const ITEM *walk;
    char *end,*next;
    int more;

    end = strchr(here,0);
    next = top ? end+3 : end+1;
    more = 0;
    for (walk = list; walk; walk = walk->next)
	if (sub_tree(walk,next,0,levels-1)) {
	    next = strchr(next,0);
	    *next++ = ',';
	    more = 1;
	}
    if (!IS_LISTABLE(curr) && !more && !top) return 0;
    if (more) {
	if (!top) *end = '(';
	else {
	    end[0] = ' ';
	    end[1] = ' ';
	    end[2] = '(';
	}
	*--next = ')';
	*++next = 0;
    }
    return 1;
}


static int sub_tree(const ITEM *curr,char *here,int top,int levels)
{
    const char *name;
    int value,first,i,summary;

    if (!levels || !curr) return 0;
    name = top || !curr->short_name ? curr->long_name : curr->short_name;
    strcpy(here,name);
    value = VALUE_OF(curr);
    if (top) {
	if (curr->type == it_set || tree_expanded(curr)) return 1;
	if (curr->type == it_title || curr->type != it_choice || value)
	    sub_list(curr,tree_section(curr),here,top,levels);
	return 1;
    }
    switch (curr->type) {
	case it_yesno:
	    if (!value) return 0;
	    /* fall through */
	case it_title:
	    break;
	case it_number:
	    if (!value) return 1;
	    sprintf(here,"%s=%d",name,value);
	    break;
	case it_choice:
	    sprintf(here,"%s=%s",name,curr->values[value].short_name ?
	      curr->values[value].short_name : curr->values[value].long_name);
	    break;
	case it_set:
	    strcpy(here,name);
	    if (levels == 1 || top) return 1;
	    first = 1;
	    for (i = 0; i < curr->num_values; i++)
		if (value & (1 << i)) {
		    strcat(here,first ? "=" : ",");
		    strcat(here,curr->values[i].short_name ?
		      curr->values[i].short_name : curr->values[i].long_name);
		    first = 0;
		    if (sub_list(curr,curr->values[i].entries,here,top,
		      levels-1)) here = strchr(here,0);
		}
	    return 1;
	default:
	    abort();
    }
    summary = sub_list(curr,tree_section(curr),here,top,levels);
    return curr->type == it_title ? summary : 1;
}


const char *tree_line(const ITEM *curr,int max_len)
{
    static char line[MAX_SCRATCH+1];
    int pick;

    if (!sub_tree(curr,line,1,-1)) return "?";
    if (strlen(line) > max_len) {
	pick = 0;
	do sub_tree(curr,line,1,++pick);
	while (strlen(line) <= max_len);
	(void) sub_tree(curr,line,1,pick-1);
    }
    return line;
}


void tree_change(ITEM *curr,int next)
{
    int max;

    assert(IS_LISTABLE(curr) || !IS_VARIABLE(curr));
    if (!IS_LISTABLE(curr)) return;
    if (curr->type == it_set) {
	curr->active = (curr->active+(next ? 1 : -1)+curr->num_values) %
	  curr->num_values;
	return;
    }
    if (curr->value == -1) curr->value = curr->deflt;
    if (curr->value == -1) curr->value = 0;
    max = tree_num_values(curr);
    curr->value = (curr->value+(next ? 1 : -1)+max) % max;
    if (curr->type == it_choice) curr->values[curr->value].expand = 1;
    if (curr->type == it_yesno && curr->value) curr->expand = 1;
}


void tree_toggle(ITEM *curr)
{
    if (curr->type != it_set) return;
    if (curr->value == -1) curr->value = curr->deflt;
    curr->value ^= 1 << curr->active;
    if (curr->value & (1 << curr->active))
	curr->values[curr->active].expand = 1;
}


void tree_set_number(ITEM *curr,int number)
{
    assert(curr->type == it_number);
    curr->value = number;
    curr->expand = 1;
}
