/* values.c - configuration data */

/* Written 1993 by Werner Almesberger */


#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include "common.h"
#include "scanner.h"
#include "values.h"
#include "descr.h"


#define MAX_CHUNK 300


char *obsolete = NULL;
static char *obs_end;
static int size;


static void expand(void)
{
    char *new;

    if (!obsolete) {
	obsolete = obs_end = alloc((size_t) (MAX_CHUNK*2));
	size = MAX_CHUNK*2;
	*obsolete = 0;
	return;
    }
    if (obsolete+size > obs_end+MAX_CHUNK) return;
    if (new = realloc(obsolete,(size_t) (size+MAX_CHUNK))) {
	obs_end = new+(obs_end-obsolete);
	obsolete = new;
	size += MAX_CHUNK;
	return;
    }
    new = alloc((size_t) (size+MAX_CHUNK));
    memcpy(new,obsolete,(size_t) (obs_end-obsolete+1));
    obs_end = new+(obs_end-obsolete);
    free(obsolete);
    obsolete = new;
}


static void add_printf(const char *fmt,...)
{
    va_list ap;

    va_start(ap,fmt);
    vsprintf(obs_end,fmt,ap);
    va_end(ap);
    obs_end = strchr(obs_end,0);
}


static void add_token(const char *token)
{
    if (!scan_quote(token)) {
	add_printf("%s",token);
	return;
    }
    *obs_end++ = '"';
    while (*token) {
	if (*token == '"' || *token == '\\') *obs_end++ = '\\';
	*obs_end++ = *token++;
    }
    *obs_end++ = '"';
    *obs_end = 0;
}


static ITEM *find_var(ITEM *curr,const char *name)
{
    ITEM *this;
    int i;

    while (curr) {
	if ((curr->name && !strcmp(curr->name,name)) || !strcmp(curr->long_name,
	  name)) break;
	if (this = find_var(curr->entries,name)) return this;
	if (curr->type == it_choice)
	    for (i = 0; i < curr->num_values; i++)
		if (this = find_var(curr->values[i].entries,name)) return this;
	curr = curr->next;
    }
    return curr;
}


static int get_item(void)
{
    ITEM *curr;
    char *tmp,*end;
    int i,value;

    curr = descr;
    while (1) {
	scan_string();
	tmp = stralloc(string);
	add_token(string);
	if (!scan_next()) {
	    token = -1;
	    break;
	}
	if (token != '.') break;
	add_printf(".");
	if (curr && (curr = find_var(curr,tmp)) && curr->type != it_set)
	    scan_error("not a set \"%s\"",tmp);
	free(tmp);
	scan_string();
	add_token(string);
	if (curr) {
	    for (i = 0; i < curr->num_values; i++)
		if (!strcmp(curr->values[i].long_name,string)) break;
		if (i == curr->num_values)
		    scan_error("no such value \"%s\"",string);
	    curr = curr->values[i].entries;
	}
	if (!scan_next()) {
	    token = -1;
	    break;
	}
	if (token != ',') break;
	add_printf(",");
    }
    if (curr) curr = find_var(curr,tmp);
    free(tmp);
    switch (token) {
	case '!':
	    if (curr && curr->type != it_yesno)
		scan_error("variable type mismatch");
	    add_printf("!");
	    if (curr) curr->value = 0;
	    break;
	case '=':
	    add_printf("=");
	    scan_string();
	    add_token(string);
	    if (curr)
		switch (curr->type) {
		    case it_number:
			value = strtol(string,&end,10);
			if (*end) scan_error("number expected");
			if (value < curr->low || value > curr->high)
			    scan_error("number out of bounds");
			curr->value = value;
			break;
		    case it_choice:
			for (i = 0; i < curr->num_values; i++)
			    if (!strcmp(curr->values[i].long_name,string))
				break;
			if (i == curr->num_values)
			    scan_error("no such value \"%s\"",string);
			curr->value = i;
			break;
		    default:
			scan_error("variable type mismatch");
		}
	    break;
	case '(':
	    if (curr && curr->type != it_set)
		scan_error("variable type mismatch");
	    add_printf("(");
	    if (curr) curr->value = 0;
	    scan_real();
	    if (token == ')') add_printf(")");
	    else {
		scan_return();
		do {
		    scan_string();
		    add_token(string);
		    if (curr) {
			for (i = 0; i < curr->num_values; i++)
			    if (!strcmp(curr->values[i].long_name,string))
				break;
			if (i == curr->num_values)
			    scan_error("no such value \"%s\"",string);
			curr->value |= (1 << i);
		    }
		    scan_real();
		    add_printf("%c",token);
		}
		while (token == ',');
	    }
	    if (token != ')') scan_error(") expected");
	    break;
	default:
	    if (token != -1) scan_return();
	    if (curr && curr->type != it_yesno)
		scan_error("variable type mismatch");
	    if (curr) curr->value = 1;
	    break;
    }
    return !!curr;
}


void val_load(const char *name)
{
    char *last;

    scan_open(name);
    while (scan_next()) {
	scan_return();
	expand();
	last = obs_end;
	if (get_item()) *(obs_end = last) = 0;
	else add_printf("\n");
    }
    scan_close();
}
